/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.merge.c,v 1.4 1998/09/21 22:21:17 muth Exp muth $
  Comment:  This file contains the code for the condbranch
            instruction merge optimization
=======================================================*/

#include "global.h"
#include "flags.h"
#include "functions.h"
#include "opcode.h"
#include "regs.h"

/*=======================================================
@FUN
Mark those instructions that have no dependencies
within the given bbl, unmark the others.
Dependecies can either be register or resources.
This functioned currently uses the default memory
disambiguation which assumes that every store conflicts
with all loads and stores.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblMarkRootIns"

static void AbblMarkRootIns(INDEX ibbl)
{
    REGLIST reg_def;
    REGLIST reg_use;
    REGLIST res_def;
    REGLIST res_use;
    
    INDEX iins;

    reg_def = REGS_NONE;
    reg_use = REGS_NONE;
    res_def = 0;
    res_use = 0;
    
    FOREACH_BBL_INS(ibbl,iins)
    {
        REGLIST curr_reg_def = AinsDef(iins);
        REGLIST curr_reg_use = AinsUse(iins);
        REGLIST curr_res_def = AINS_RES_DEF(iins);
        REGLIST curr_res_use = AINS_RES_USE(iins);
        
        if( (curr_reg_use & reg_def) ||
            (curr_reg_def & (reg_def|reg_use)) ||
            (curr_res_use & res_def) ||
            (curr_res_def & (res_def|res_use)) )
        {
            AINS_FLAGS(iins) &= ~F_MARKED;
        }
        else
        {
            AINS_FLAGS(iins) |= F_MARKED;
        }

        reg_use |= curr_reg_use;
        reg_def |= curr_reg_def;
        res_use |= curr_res_use;
        res_def |= curr_res_def;
    }
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsEqual"

static BOOL AinsEqual(INDEX iins,INDEX jins)
{
    if( AINS_CODE(iins) != AINS_CODE(jins) ) return 0;
    if( AINS_REG_C(iins) != AINS_REG_C(jins) ) return 0;
    if( AINS_REG_B(iins) != AINS_REG_B(jins) ) return 0;
    if( AINS_REG_A(iins) != AINS_REG_A(jins) ) return 0;
    if( AINS_DATA(iins) != AINS_DATA(jins) ) return 0;

    return 1;
}

/*=======================================================
@OPT
This optimization tries to move instructions which appear
both in the true and the false branch of a conditional
before that conditional. It does this by determining a set
of movable instructions in both basic blocks
and the checking for identical instructions in those sets.
Once an instruction has been moved the set of movable instructions
might change, so we restart the process.

Requiers: Liveness

Benefits from Register Renaming (to reduce the number of registers used)
since this often normalizes the register names it also provides
free registers for renaming activities
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptMergeInsUp"

extern BOOL OptMergeInsUp(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Mergeinsup");
    BOOL   change;
    INDEX  ibbl;
    INDEX  ibbl_t;
    INDEX  ibbl_f;
    INDEX  iins_branch;
    INDEX  iins_t;
    INDEX  iins_f;
    REG    reg_condition;
    
    NUMBER count = 0;
    NUMBER counth = 0;
    NUMBER countr = 0;
    NUMBER countp = 0;
    NUMBER countph = 0;
        
    clock_t time = clock();
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_BRANCH ) continue;
        
        iins_branch = ABBL_INS_LAST(ibbl);
        reg_condition = AINS_REG_A(iins_branch);
        
        ibbl_t = AEDG_TAIL(ABBL_EDG_TRUE(ibbl));
        ibbl_f = AEDG_TAIL(ABBL_EDG_FALSE(ibbl));

        if( AbblNpreds(ibbl_t) != 1 ) continue;
        if( AbblNpreds(ibbl_f) != 1 ) continue;

    restart:
        AbblMarkRootIns(ibbl_t);
        AbblMarkRootIns(ibbl_f);

        FOREACH_BBL_INS(ibbl_t,iins_t)
        {
            if( !(AINS_FLAGS(iins_t) & F_MARKED) ) continue;
            
            FOREACH_BBL_INS(ibbl_f,iins_f)
            {
                if( !(AINS_FLAGS(iins_f) & F_MARKED) ) continue;

                if( !AinsEqual(iins_t,iins_f) ) continue;

                if( AINS_TYPE(iins_t) == IT_JMP ||
                    AINS_TYPE(iins_t) == IT_BRA ||
                    AINS_TYPE(iins_t) == IT_IBR ||
                    AINS_TYPE(iins_t) == IT_FBR ||
                    AINS_TYPE(iins_t) == IT_PAL )
                {
                    if( verbose )
                    {
                        POTENTIAL(("Optimization of branches not implemented\n"));
                        PrintIns(stdout,iins_t);
                        PrintIns(stdout,iins_f);
                        PrintBblSuper(stdout,ibbl);
                        PrintBblSuper(stdout,ibbl_t);
                        PrintBblSuper(stdout,ibbl_f);
                    }
                    continue;
                }

                countp++;
                if( ABBL_IS_HOT(ibbl) ) countph++;
                
                if( AINS_REG_C(iins_t) == reg_condition )
                {
                    REGLIST regs_avail;
                    INDEX iins_def;
                    REG newreg;
                    
                    if( AinsUse(iins_t) & REG2BIT(reg_condition) ) continue;
                    
                    CERROR( ABBL_OUT(ibbl) & REG2BIT(reg_condition),("condition register should be dead\n"));

                    if( ABBL_OUT(ibbl) & REG2BIT(reg_condition) )
                    {
                        PrintIns(stdout,iins_t);
                        PrintBblSuper(stdout,ibbl);
                        PrintBblSuper(stdout,ibbl_f);
                        PrintBblSuper(stdout,ibbl_t);
                    }
                    
                    iins_def = FindInstructionRegDef(iins_branch, reg_condition);

                    if( !VALID(iins_def) ) continue;
                    
                    regs_avail =  MyFindScratchRegs(iins_def,INVALID);
                    
                    regs_avail &= REG_IS_INT(reg_condition) ?  REGS_INT_BUT_ZERO : REGS_FLT_BUT_ZERO;
                        
                    if( regs_avail == REGS_NONE ) continue;

                    newreg = RegsGetFirst(regs_avail);
                    
                    if( verbose )
                    {
                        printf("\t[*] renaming condition register %d -> %d\n",reg_condition,newreg);
                    }

                    reg_condition = newreg;
                    
                    RenameRegister(iins_def,newreg);
                    countr++;
#if 0
                    PrintBblSuper(stdout,ibbl);
#endif
                    
                }
                    
                count++;
                if( ABBL_IS_HOT(ibbl) ) counth++;
                
                if( verbose && ABBL_COUNT(ibbl) > 0 )
                {
                    HURRAH(("Merging\n"));
                    PrintIns(stdout,iins_t);
                    PrintIns(stdout,iins_f);
                    PrintBblSuper(stdout,ibbl);
                    PrintBblSuper(stdout,ibbl_t);
                    PrintBblSuper(stdout,ibbl_f);
                }

                                  
                AinsKill(iins_t);
                iins_t = AinsDup(iins_f);
                AINS_BBL(iins_t) = ibbl;
                AinsKill(iins_f);
                AinsInsertBefore(iins_t,iins_branch);

                /* reconstruct the liveness information */

                ABBL_OUT(ibbl) = AbblRegsLiveIn(ibbl_f) | AbblRegsLiveIn(ibbl_t);
                
                goto restart;
            }
            
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = (count>0);

    InfoStatus(fp,"\t[OptMergeinsup%c] merged iins: %d#%d potential: %d#%d  regs renamed: %d {%d ms}\n",
            CHANGE(change),count,counth,countp,countph,countr,time);
    
    return 0;   /* there should be only one run of this optimization */
}
/*=======================================================
=======================================================*/
