/* ---------------------------------------------------------------
**  This file builds the tables used by the instruction scheduler:
**
**     ProdConsLatency[<producer>][<consumer>]
**
**     ProdProdLatency[<first producer>][<second producer>]
** ---------------------------------------------------------------
*/

#include "global.h"
#include "functions.h"
#include "schedglob.h"
#include "flags.h"

/* ----------------------------------------------------------------- */

#define SCH_CL( Name, PL1, PL2, Lat1, Lat2, MulAdd, Special ) \
    { #Name, (PL_ ## PL1 | PL_ ## PL2) , Lat1, Lat2, MulAdd, Special },

SCH_BASIC_TABLE_ENTRY SchBasicTable[] = {
#include "schedclasses.tbl"
   {"LAST ENTRY"}};

#undef SCH_CL

/* ----------------------------------------------------------------- */

LATENCY ProdConsLatency [NO_OF_SCHED_CLASSES] [NO_OF_SCHED_CLASSES];
LATENCY ProdProdLatency [NO_OF_SCHED_CLASSES] [NO_OF_SCHED_CLASSES];
LATENCY FuncUnitUse     [NO_OF_FUNC_UNITS   ] [NO_OF_SCHED_CLASSES];

/* ----------------------------------------------------------------- */

#undef __PROC__
#define __PROC__ "AINS_SCHEDCLASS"

SCHED_CLASS AINS_SCHEDCLASS(INDEX iins)
{
    switch( AINS_TYPE(iins) )
    {
    case IT_F2R:
        return SC_MFPR;
        
    case IT_R2F:
        return SC_MTPR;

    case IT_MEM:
        return SC_MBX2;
        
    case IT_LLD:
    case IT_CST:
        return SC_MBX1;
        
    case IT_MSC:
        return SC_MISC;
        
    case IT_ILD:
    case IT_FLD:
    case IT_ULD:
        return SC_LD;

    case IT_IST:
    case IT_FST:
    case IT_UST:
        return SC_ST;

    case IT_LDA:
    case IT_IOP:
        return SC_IADD;

    case IT_FCM:
    case IT_FCP:
    case IT_FOP:
        return SC_FADD;
        
    case IT_IBR:
        return SC_IBR;
        
    case IT_FBR:
        return SC_FBR;

    case IT_BRA:

    case IT_PAL:

    case IT_JMP:

        return SC_JSR;

    case IT_ICM:
        return SC_CMOV;
       
    case IT_ICP:
        return SC_ICMP;
    
    case IT_FM4:
    case IT_FM8:
        return SC_FMUL;
    
    case IT_FD4:
        return SC_FDIV1;
        
    case IT_FD8:
        return SC_FDIV2;
     
    case IT_LOG:
        return SC_ILOG;
        
    case IT_SFT:
        return SC_SHIFT;
    
    case IT_MLL:
        return SC_IMULL;
            
    case IT_MLQ:
        return SC_IMULQ;
        
    case IT_MLH:
        return SC_IMULH;

    case IT_RX:
        return SC_RX;
        
    case IT_FCY:
        return SC_FCPYS;
    
    case IT_EXT:
        return SC_SEXT;

    case IT_F2I:
    case IT_I2F:
    default:
        ERROR(("unexpected instruction type\n"));
        return INVALID;
    }
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "MakeLatencyTableProdCons"

extern void MakeLatencyTableProdCons()
{
   SCHED_CLASS producer;
   SCHED_CLASS consumer;
   LATENCY     latency;

   FOREACH_SCHED_CLASS(producer)
   {
      FOREACH_SCHED_CLASS(consumer)
      {
	 if ( SCHCL_IS_SPECIAL(producer) )
	    {
	       latency = NoLat;  /* The cases in the switch below where
				    there is no assignment for latency
				    should never apply.  But just in case,
				    we do this assignment */
	       
	       switch(producer) {
	       case SC_LD:

	       /*
	       **  Latency varies depending on cache hit/miss.
	       **  We are optimistic and use the chache hit latency.
	       */

		  latency = SCHCL_LATENCY1(producer);
		  break;

	       case SC_ST:
		  break;

	       case SC_MBX1:

	       /*
	       **  latency depends on memory subsystem state
	       */

		  latency = 0;
		  break;

	       case SC_MBX2:
		  break;

	       case SC_RX:
		  break;

	       case SC_MTPR:
		  break;

	       case SC_MFPR:
			
	       /*
	       **  latency is either latency1 or latency2 or longer			   
	       **  (depends on the IPR) where latency1 < latency2.
	       **  Lets pick latency2.
	       */
			
		  latency = SCHCL_LATENCY2(producer);
		  break;

	       case SC_IBR:
	       case SC_FBR:
	       case SC_JSR:

	       /*
	       **  It may seem that there should be no result of a
	       **  branch or jump and thus no producer/consumer latency.
	       **  This is probably true most of the time but there is at
	       **  least one exception in that the instruction call_pal
	       **  (wich belongs to JSR) can occur other places in a
	       **  basic block than the last and may produce results(?).
	       */

		  latency = SCHCL_LATENCY1(producer);
		  break;

	       case SC_IADD:
		  break;

	       case SC_ILOG:

	       /*
	       **  less latency (actualy none) when result is
	       **  used by IBR or CMOV.
	       */

		  if (consumer == SC_IBR || consumer == SC_CMOV)
		     {
			latency = SCHCL_LATENCY2(producer);
		     }
		  else
		     {
			latency = SCHCL_LATENCY1(producer);
		     }
		  break;

	       case SC_SEXT:
		  break;

	       case SC_SHIFT:
		  break;

	       case SC_CMOV:
		  break;

	       case SC_ICMP:

	       /*
	       **  less latency (actualy none) when result is
	       **  used by IBR or CMOV.
	       */

		  if (consumer == SC_IBR || consumer == SC_CMOV)
		     {
			latency = SCHCL_LATENCY2(producer);
		     }
		  else
		     {
			latency = SCHCL_LATENCY1(producer);
		     }
		  break;

	       case SC_IMULL:
	       case SC_IMULQ:
	       case SC_IMULH:

	       /*
	       **  Use latency 2 if the consumer is also an IMUl*
	       **  instruction.
	       */

		  if (consumer == SC_IMULL ||
		      consumer == SC_IMULQ ||
		      consumer == SC_IMULH )
		     {
			latency = SCHCL_LATENCY2(producer);
		     }
		  else
		     {
			latency = SCHCL_LATENCY1(producer);
		     }
		  break;

	       case SC_FADD:
		  break;
		  
	       case SC_FDIV1:
	       case SC_FDIV2:

	       /*
	       **  Floating point divide has a min and a max latency.
	       **  We pick the max latency.
	       */

		  latency = SCHCL_LATENCY2(producer);
		  break;

	       case SC_FMUL:
		  break;

	       case SC_FCPYS:
		  break;

	       case SC_MISC:
		  break;
	       case SC_UNOP:
		  break;
	       }
	    }
	 else
	    {
	       latency = SCHCL_LATENCY1(producer);
	    }

	 /*
	 **  There is an additional latency for an IMUL* instruction.
	 **  This additional latency depends on the classes of the
	 **  instuctions that produces its input.
	 **
	 **  This will not fit in our two dimensional scheme.  The
	 **  additional latency will therfore be added to the
	 **  instruction that produces the input to an IMUL* instruction
	 **  instead.  According to remark 3 in Table 2-9 this should be
	 **  OK.
	 */

	 if ( latency != NoLat &&
	     (consumer == SC_IMULL ||
	      consumer == SC_IMULQ ||
	      consumer == SC_IMULH   ))
	    {
	       latency += SCHCL_MULADD(producer);
	    }

	 ProdConsLatency[producer][consumer] = latency;

      }
   }
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "MakeLatencyTableProdProd"

extern void MakeLatencyTableProdProd()
{
   SCHED_CLASS producer1;
   SCHED_CLASS producer2;

   /*
   **  From the Hardware Manual:
   **
   **    "Producer--Producer latency, also known as write-after-write
   **    conflicts, case issue-stalls to preserve write order.  If two
   **    instructions write the same registers, they are forced to do
   **    so in different cycles by the IDU. [...]  For most instructions,
   **    the order in which they write the register file is dictated by
   **    issue order."
   **
   **  It seems therefore that it would be a good idea to have  at least
   **  a latency of one cycle between instructions writing the same
   **  register.
   **
   **  NOTE: in the DEC scheduler they seem to have a zero cycle latency
   **        as default.  Why?
   */

   FOREACH_SCHED_CLASS(producer1)
   {
      FOREACH_SCHED_CLASS(producer2)
      {
	 ProdProdLatency[producer1][producer2] = 1;
      }
   }

   /*
   **  But some instructions should be treated specially:
   **
   **  From the Hardware Manual:
   **
   **    "... IMUL, FDIV and LD instruction may require more time
   **    than other instruction to complete."
   **
   **  and
   **
   **    "If a load writes a register, and within two cycles [The
   **    Dcache hit latency of LD] a subsequent instruction writes
   **    the same register, the subsequent instructon is issued
   **    speculatively, assuming the load hits."
   **
   **  Thus, for IMUL and FDIV we use their producer/consumer
   **  latency and for LD we keep the one cycle default latency.
   */

  FOREACH_SCHED_CLASS(producer2)
  {
     ProdProdLatency[SC_IMULL][producer2] =
	ProdConsLatency[SC_IMULL][producer2];
  }
   
  FOREACH_SCHED_CLASS(producer2)
  {
     ProdProdLatency[SC_IMULQ][producer2] =
	ProdConsLatency[SC_IMULQ][producer2];
  }
   
  FOREACH_SCHED_CLASS(producer2)
  {
     ProdProdLatency[SC_IMULH][producer2] =
	ProdConsLatency[SC_IMULH][producer2];
  }
   
  FOREACH_SCHED_CLASS(producer2)
  {
     ProdProdLatency[SC_FDIV1][producer2] =
	ProdConsLatency[SC_FDIV1][producer2];
  }

  FOREACH_SCHED_CLASS(producer2)
  {
     ProdProdLatency[SC_FDIV2][producer2] =
	ProdConsLatency[SC_FDIV2][producer2];
  }
   
}


/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "MakeLatencyTableFuncUnitUse"

extern void MakeLatencyTableFuncUnitUse()
{
   FUNC_UNIT    fu;
   SCHED_CLASS  sc;

   /*
   **  Most instructions does not hold up any functional units.  The
   **  exceptions are IMULx and FDIV.
   */

   FOREACH_FUNC_UNIT(fu)
   {
      FOREACH_SCHED_CLASS(sc)
      {
	 FuncUnitUse[fu][sc] = 0;
      }
   }

   FuncUnitUse[FU_IMULTIPLIER][SC_IMULL] = SCHCL_LATENCY2(SC_IMULL);
   FuncUnitUse[FU_IMULTIPLIER][SC_IMULQ] = SCHCL_LATENCY2(SC_IMULQ);
   FuncUnitUse[FU_IMULTIPLIER][SC_IMULH] = SCHCL_LATENCY2(SC_IMULH);
   FuncUnitUse[FU_FDIVIDER   ][SC_FDIV1] = SCHCL_LATENCY1(SC_FDIV1);
   FuncUnitUse[FU_FDIVIDER   ][SC_FDIV2] = SCHCL_LATENCY1(SC_FDIV2);
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "PrintLatencyTableHead"

void PrintLatencyTableHead(int RowLabelWidth)
{
   int i = 0;
   int j;
   SCHED_CLASS sch_cl;

   printf("/*\n");

   FOREACH_SCHED_CLASS(sch_cl)
   {
      printf("%*s", 8+RowLabelWidth, " ");

      for( j=0; j<i; j++ )
	 printf("|   ");
      i++;

      printf("%s\n", SCHCL_NAME(sch_cl));
   }

   printf("%*s---", 5+RowLabelWidth, " ");

   FOREACH_SCHED_CLASS(sch_cl)
   {
      printf("----");
   }

   printf(" */\n");


}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "PrintLatencyTableRow"

void PrintLatencyTableRow(LATENCY LatencyTableRow[NO_OF_SCHED_CLASSES],
			SCHED_CLASS from )
{
   SCHED_CLASS to;

   printf(" /* %-5s */", SCHCL_NAME(from) );

   if (from == FIRST_SCHED_CLASS)
      printf("{{ ");
   else
      printf(" { ");

   for(to = FIRST_SCHED_CLASS;
       to < LAST_SCHED_CLASS;
       to++)
   {
      printf("%2d, ", LatencyTableRow[to]);
   }

   printf("%2d", LatencyTableRow[LAST_SCHED_CLASS]);

   if (from == LAST_SCHED_CLASS)
      printf("}}\n");
   else
      printf("},\n");
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "PrintLatencyTable"

extern void PrintLatencyTable(LATENCY LatencyTable[NO_OF_SCHED_CLASSES][NO_OF_SCHED_CLASSES])
{
   SCHED_CLASS from;

   PrintLatencyTableHead(8);

   FOREACH_SCHED_CLASS(from)
   {
      PrintLatencyTableRow(LatencyTable[from], from);
   }
}


/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "PrintFuncUnitUseRow"

void PrintFuncUnitUseRow(LATENCY FUUTableRow[NO_OF_SCHED_CLASSES],
			   FUNC_UNIT fu )
{
   SCHED_CLASS sc;
   STRING      row_label;
   
   switch (fu) {
   case  FU_IMULTIPLIER:
      row_label = "IMULTIPLIER";
      break;
   case FU_FDIVIDER:
      row_label = "FDIVIDER";
      break;
   default:
      row_label = " -- ??? --";
      break;
   }

   printf(" /* %-15s */", row_label);

   if (fu == FIRST_FUNC_UNIT)
      printf("{{ ");
   else
      printf(" { ");

   for(sc = FIRST_SCHED_CLASS;
       sc < LAST_SCHED_CLASS;
       sc++)
   {
      printf("%2d, ", FUUTableRow[sc]);
   }

   printf("%2d", FUUTableRow[LAST_SCHED_CLASS]);

   if (fu == LAST_FUNC_UNIT)
      printf("}}\n");
   else
      printf("},\n");
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "PrintFuncUnitUseTable"

void PrintFuncUnitUseTable(LATENCY FUUTable[NO_OF_FUNC_UNITS]
			                   [NO_OF_SCHED_CLASSES])
{
   FUNC_UNIT fu;

   PrintLatencyTableHead(18);

   FOREACH_FUNC_UNIT(fu)
   {
      PrintFuncUnitUseRow(FUUTable[fu], fu);
   }
}


/*=====================================================================
  =====================================================================*/
#if 0

#undef __PROC__
#define __PROC__ "main"

void main()
{
   MakeLatencyTableProdCons();
   MakeLatencyTableProdProd();
   MakeLatencyTableFuncUnitUse();

   printf("/* Producer/Consumer Latency Table \n"
	  "   ------------------------------- \n"
	  "   This file has been generated automatically from ...\n"
	  "*/\n");

   PrintLatencyTable(ProdConsLatency);
   PrintLatencyTable(ProdProdLatency);
   PrintFuncUnitUseTable(FuncUnitUse);
}
#endif
